#include <string.h>
#include <stdio.h>
#include <iostream.h>

#include "strm.h"
#include "util.h"

#define DebugSeek 0

template<class T> FitsStream<T>::FitsStream()
{
  seek_ =0;
  stream_ =0;
}

template<class T> FitsStream<T>::~FitsStream()
{
  if (data_)
    delete (T*)data_;
}

// FILE*

int FitsStream<FILE*>::read(char* where, int size)
{
  return fread(where, 1, size, stream_);
}

int FitsStream<FILE*>::tell()
{
  return ftell(stream_);
}

int FitsStream<FILE*>::seek(int size)
{
  return fseek(stream_, size, SEEK_CUR);
}

void FitsStream<FILE*>::close()
{
  fclose(stream_);
}

// Tcl_Channel

int FitsStream<Tcl_Channel>::read(char* where, int size)
{
  return Tcl_Read(stream_, where, size);
}

int FitsStream<Tcl_Channel>::tell()
{
  return Tcl_Tell(stream_);
}

int FitsStream<Tcl_Channel>::seek(int size)
{
  return Tcl_Seek(stream_, size, SEEK_CUR);
}

void FitsStream<Tcl_Channel>::close()
{
  // do nothing here
}

template<class T> FitsHead* FitsStream<T>::headRead()
{
  // read first block

  char* cards = new char[FT_BLOCK];
  if (!cards)
    return NULL;

  memset(cards, ' ', FT_BLOCK);
  if (read(cards, FT_BLOCK) != FT_BLOCK) {
    delete [] cards;
    return NULL;
  }

  // simple FITS file check

  if (strncmp(cards, "SIMPLE  =", 9) && strncmp(cards, "XTENSION=", 9)) {
    delete [] cards;
    return NULL;
  }

  // read remaining blocks

  int numblks = 1;
  char* current = cards;
  while (1) {
    if (findEnd(current))
      break;

    // this is check
    // for TCL channels, the stream gets corrupted, so END is never found
    // so bail out after an extreme number of blocks have been read

    if (numblks>10) {
      delete [] cards;
      return NULL;
    }

    char* tmp = new char[(numblks+1)*FT_BLOCK];
    memcpy(tmp, cards, numblks*FT_BLOCK);
    delete [] cards;
    cards = tmp;
    current = cards + numblks*FT_BLOCK;
    memset(current, ' ', FT_BLOCK);

    if (read(current, FT_BLOCK) != FT_BLOCK) {
      delete [] cards;
      return NULL;
    }
    
    numblks++;
  }

  // create header

  FitsHead* fits = new FitsHead(cards, numblks*FT_BLOCK, FitsHead::ALLOC);
  if (!fits->isValid()) {
    delete fits;
    return NULL;
  }

  seek_ = tell();
  if (DebugSeek)
    cerr << "headRead: " << seek_ << endl;

  return fits;
}

template<class T> void* FitsStream<T>::dataRead(int bytes)
{
  if (bytes) {
    char* data = new char[bytes];
    if (!data)
      return NULL;

    if (read(data, bytes) != bytes) {
      delete [] data;
      return NULL;
    }

    seek_ = tell();
    if (DebugSeek)
      cerr << "dataRead: " << seek_ << endl;

    return data;
  }
}

template<class T> void FitsStream<T>::dataSkip(int bytes)
{
  if (bytes) {
    if (canSeek()) {
      seek(bytes);
      seek_ = tell();
      if (DebugSeek)
	cerr << "dataSkip: " << seek_ << endl;
      return;
    }
    else {
      char block[FT_BLOCK];

      while (bytes > 0) {
	read(block, (bytes < FT_BLOCK ? bytes : FT_BLOCK));
	bytes -= FT_BLOCK;
      }
    }
  }
}

template<class T> void FitsStream<T>::processExact()
{
  seek_ = tell();
  if (DebugSeek)
    cerr << "processExact: " << seek_ << endl;

  if (!(pExt_ || pIndex_ > 0)) {

    // we are only looking for a primary image

    if (head_ = headRead())
      goto found;
  }
  else {

    // we are looking for an extension
    // keep the primary header

    if (!(primary_ = headRead()))
      goto error;

    if (pExt_) {
      while (1) {
	if (!(head_ = headRead()))
	  goto error;

	if (head_->extname()) {
	  char* a = fttoupper(head_->extname());
	  char* b = fttoupper(pExt_);
	  if (!strcmp(a,b)) {
	    delete [] a;
	    delete [] b;
	    goto found;
	  }
	  delete [] a;
	  delete [] b;
	}

	dataSkip(head_->datablocks()*FT_BLOCK);
	delete head_;
	head_ = NULL;
      }
    }
    else {
      for (int i=1; i<pIndex_; i++) {
	if (!(head_ = headRead()))
	  goto error;

	dataSkip(head_->datablocks()*FT_BLOCK);
	delete head_;
	head_ = NULL;
      }

      if (head_ = headRead())
	goto found;
    }
  }

  // Must have an error

error:
  if (head_)
    delete head_;
  head_ = NULL;
  if (primary_)
    delete primary_;
  primary_ = NULL;

  data_ = NULL;
  valid_ = 0;
  close();
  return;

  // found it!

found:
  data_ = dataRead(head_->databytes());
  valid_ = 1;
  close();
  return;
}

template<class T> void FitsStream<T>::processRelax()
{
  seek_ = tell();
  if (DebugSeek)
    cerr << "processRelax: " << seek_ << endl;

  // check to see if there is an image in the primary

  head_ = headRead();
  if (!head_)
    goto error;
  else {
    if (head_->isValid() && head_->naxes() > 0)
      goto found;
  }

  // ok, no image, save primary and lets check extensions

  primary_ = head_;
  dataSkip(head_->datablocks()*FT_BLOCK);
  head_ = NULL;

  // ok, no image, lets check extensions

  while (1) {
    if (!(head_ = headRead()))
      goto error;

    if (head_->isImage() && head_->naxes() > 0)
      goto found;

    if (head_->isTable() && head_->extname()) {
      char* a = fttoupper(head_->extname());
      if ((!strcmp("STDEVT", a)) || !strcmp("EVENTS", a)) {
	delete [] a;
	goto found;
      }
      else
	delete [] a;
    }

    dataSkip(head_->datablocks()*FT_BLOCK);
    delete head_;
    head_ = NULL;
  }

  // did not find anything, bail out

 error:
  if (head_)
    delete head_;
  head_ = NULL;
  if (primary_)
    delete primary_;
  primary_ = NULL;

  data_ = NULL;
  valid_ = 0;
  close();
  return;

  // found it!

 found:
  data_ = dataRead(head_->databytes());
  valid_ = 1;
  close();
  return;
}

template<class T> void FitsStream<T>::processArray()
{
  // check to see if we have a nonzero width, height, and bitpix
  if (!validArrayParams())
    goto error;

  // skip header
  if (pSkip_)
    dataSkip(pSkip_);

  // read data
  data_ = dataRead(pWidth_ * pHeight_ * abs(pBitpix_)/8);
  if (!data_)
    goto error;

  // create blank header
  head_ = new FitsHead(pWidth_, pHeight_, pBitpix_);
  if (!head_->isValid())
    goto error;

  // do we need to byteswap?
  setArrayByteSwap();

  // made it this far, must be good
  valid_ = 1;
  close();
  return;

error:
    valid_=0;
    close();
    return;
}

template<class T> void FitsStream<T>::processMosaic()
{
  primary_ = headRead();
  if (!(primary_ && primary_->isValid()))
    goto error;

  dataSkip(primary_->datablocks()*FT_BLOCK);

  // first extension

  head_ = headRead();
  if (!(head_ && head_->isValid()))
    goto error;

  // be sure to read all blocks, so that the next call starts on a boundary
  data_ = dataRead(head_->datablocks()*FT_BLOCK);
  valid_ = 1;
  return;

  // Must have an error

error:
  if (head_)
    delete head_;
  head_ = NULL;
  if (primary_)
    delete primary_;
  primary_ = NULL;

  data_ = NULL;
  valid_ = 0;
  close();
  return;
}

template<class T> void FitsStream<T>::processMosaicNext(FitsFile* p)
{
  FitsStream<T>* prev = (FitsStream<T>*)p;
  primary_ = prev->primary();
  stream_ = prev->stream();
  managePrimary_ = 0;

  head_ = headRead();
  if (!(head_ && head_->isValid())) {
    if (head_)
      delete head_;
    head_ = NULL;

    data_ = NULL;
    valid_ = 0;
    close();
    return;
  }

  // be sure to read all blocks, so that the next call starts on a boundary

  data_ = dataRead(head_->datablocks()*FT_BLOCK);
  valid_ = 1;
  return;
}

template class FitsStream<FILE*>;
template class FitsStream<Tcl_Channel>;
