#include "ellipse.h"
#include "frame.h"
#include "util.h"

Ellipse::Ellipse(Frame* p, const Vector& ctr, const Vector& r, double a, 
		 const char* clr,  const char* f, const char* t, 
		 unsigned short prop)
  : AnnuliMarker(p, ctr, 48, clr, f, t, prop) 
{
  radius = r;
  angle = a;
  strcpy(type,"ellipse");
  handle = new Vector[4];
  numHandle = 4;

  updateBBox();
}

void Ellipse::updateCoords(Matrix& m)
{
  radius *= Scale(m);
  Marker::updateCoords(m);
}

void Ellipse::edit(const Vector& v, int h)
{
  radius = ((v-center)*Rotate(angle)).abs();

  updateBBox();
  doCallBacks(&editCB);
}

void Ellipse::rotate(const Vector& v, int h)
{
  // v is in ref coords
  // handles are in canvas coords

  Vector a = v - center;
  Vector b = (handle[h-1] * parent->canvasToRef) - center;

  angle += a.angle() - b.angle();

  updateBBox();
  doCallBacks(&rotateCB);
}

int Ellipse::isIn(const Vector& v)
{
  // v is in canvas coords

  Vector r = radius * maxAnnuli();
  if (properties & FIXED)
    r /= parent->getZoom();

  Vector p = v * parent->canvasToRef * Translate(-center) * Rotate(angle);

  if ((p[0]*p[0])/(r[0]*r[0]) + (p[1]*p[1])/(r[1]*r[1]) <= 1)
    return 1;
  else
    return 0;
}

void Ellipse::setRadius(const Vector& r)
{
  radius = r;
  updateBBox();

  doCallBacks(&editCB);
}

void Ellipse::setAngle(double a)
{
  angle = a;
  updateBBox();

  doCallBacks(&rotateCB);
}

// private

void Ellipse::calcPoints(double zoom)
{
  Matrix m;
  if (properties & FIXED)
    m = Scale(1/zoom) * Rotate(-angle) * Translate(center);
  else
    m = Rotate(-angle) * Translate(center);

  for (int i=0; i<annuli; i++) {
    Matrix m2 = Scale(ratio[i]) * m;
    for (int j=0; j<segments; j++) {
      double theta = 2*M_PI / segments * j;
      vertices[i][j] = Vector(radius[0]*cos(theta), radius[1]*sin(theta)) * m2;
    }
  }
}

void Ellipse::updateHandles()
{
  Vector r = radius * maxAnnuli();
  if (properties & FIXED)
    r /= parent->getZoom();

  Matrix m = Rotate(-angle) * Translate(center) * parent->refToCanvas;
  handle[0] = Vector(-r[0],-r[1]) * m;
  handle[1] = Vector( r[0],-r[1]) * m;
  handle[2] = Vector( r[0], r[1]) * m;
  handle[3] = Vector(-r[0], r[1]) * m;
}

// list

void Ellipse::list(ostream& str, int sys, int format)
{
  switch ((Frame::CoordSystem)sys) {
  case Frame::IMAGE:
  case Frame::PHYSICAL:
  case Frame::DETECTOR:
  case Frame::AMPLIFIER:
    {
      Vector r = parent->mapDistFromRef(radius,(Frame::CoordSystem)sys,
					Frame::PIXELS);
      Vector v = parent->map(center,Frame::REF,(Frame::CoordSystem)sys);
      str << type << setprecision(8) << v << r << radToDeg(angle);
    }
    break;

  case Frame::EQUATORIAL_FK4:
  case Frame::EQUATORIAL_FK5:
  case Frame::ECLIPTIC:
  case Frame::GALACTIC:
  case Frame::LINEAR:
  case Frame::ICRS:
    Vector r = parent->mapDistFromRef(radius, 
				      (Frame::CoordSystem)sys,Frame::ARCSEC);
    switch ((Frame::CoordFormat)format) {
    case Frame::DEGREES:
      {
	Vector v = parent->map(center,Frame::REF,(Frame::CoordSystem)sys);
	str << type << setprecision(8) << v << r[0] << "\" " << r[1] << "\" "
	    << radToDeg(angle);
      }
      break;
    case Frame::SEXAGESIMAL:
      {
	char buf[64];
	parent->mapToWCSStr(center, Frame::REF, (Frame::CoordSystem)sys,
			    (Frame::CoordFormat)format, buf, 64);
	char ra[16];
	char dec[16];
	istrstream wcs(buf,64);
	wcs >> ra >> dec;
	str << type << ' ' << ra << ' ' << dec << ' ' 
	    << r[0] << "\" " << r[1] << "\" "
	    << radToDeg(angle);
      }
      break;
    }
    break;
  }

  listProperties(str);
}

void Ellipse::listSAOtng(ostream& str, int sys, int format)
{
  if (text[0])
    str << '#' << text << endl;

  if (properties&INCLUDE)
    str << '+';
  else
    str << '-';

  // radius is always in image coords

  switch ((Frame::CoordSystem)sys) {
  case Frame::IMAGE:
  case Frame::DETECTOR:
  case Frame::AMPLIFIER:
    {
      Vector v = parent->map(center,Frame::REF,Frame::IMAGE);
      str << type << '(' << setprecision(8) << v[0] << ',' << v[1] << ',' 
	  << radius[0] << ',' << radius[1] << ',' << radToDeg(angle) << ") ";
    }
    break;
  case Frame::PHYSICAL:
    {
      Vector r = parent->mapDistFromRef(radius,
					(Frame::CoordSystem)sys,Frame::PIXELS);
      Vector v = parent->map(center,Frame::REF,(Frame::CoordSystem)sys);
      str << type << '(' << setprecision(8) << v[0] << ',' << v[1] << ',' 
	  << r[0] << ',' << r[1] << ',' << radToDeg(angle) << ") ";
    }
    break;
  case Frame::EQUATORIAL_FK4:
  case Frame::EQUATORIAL_FK5:
  case Frame::ECLIPTIC:
  case Frame::GALACTIC:
  case Frame::LINEAR:
  case Frame::ICRS:
    switch ((Frame::CoordFormat)format) {
    case Frame::DEGREES:
      {
	Vector v = parent->map(center,Frame::REF,(Frame::CoordSystem)sys);
	str << type << '(' << setprecision(8) << v[0] << ',' << v[1] << ','
	    << radius[0] << ',' << radius[1] << ',' << radToDeg(angle) << ") ";
      }
      break;
    case Frame::SEXAGESIMAL:
      {
	char buf[64];
	parent->mapToWCSStr(center, Frame::REF, (Frame::CoordSystem)sys,
			    (Frame::CoordFormat)format, buf, 64);
	char ra[16];
	char dec[16];
	istrstream wcs(buf,64);
	wcs >> ra >> dec;
	str << type << '(' << ra << ',' << dec << ',' 
	    << radius[0] << ',' << radius[1] << ',' << radToDeg(angle) << ") ";
      }
      break;
    }
    break;
  }

  str << "# " << colorName;
}

void Ellipse::listPros(ostream& str, int sys, int format)
{
  switch ((Frame::CoordSystem)sys) {
  case Frame::IMAGE:
  case Frame::DETECTOR:
  case Frame::AMPLIFIER:
    {
      Vector r = parent->mapDistFromRef(radius,(Frame::CoordSystem)sys,
					Frame::PIXELS);
      Vector v = parent->map(center,Frame::REF,Frame::IMAGE);
      str << type << ' ' << setprecision(8) << v << r << radToDeg(angle);
    }
    break;

  case Frame::PHYSICAL:
    {
      str << "physical; ";
      Vector r = parent->mapDistFromRef(radius,(Frame::CoordSystem)sys,
					Frame::PIXELS);
      Vector v = parent->map(center,Frame::REF,(Frame::CoordSystem)sys); 
      str << type << ' ' << setprecision(8) << v << r << radToDeg(angle);
    }
    break;

  case Frame::EQUATORIAL_FK4:
  case Frame::EQUATORIAL_FK5:
  case Frame::ECLIPTIC:
  case Frame::GALACTIC:
  case Frame::LINEAR:
  case Frame::ICRS:
    parent->listProsCoordSystem(str, (Frame::CoordSystem)sys);

    // display wcs radius in ARCSEC
    Vector r = parent->mapDistFromRef(radius,(Frame::CoordSystem)sys,
				      Frame::ARCSEC);

    str << "; ";
    switch ((Frame::CoordFormat)format) {
    case Frame::DEGREES:
      {
	Vector v = parent->map(center,Frame::REF,(Frame::CoordSystem)sys);
	str << type << ' ' << setprecision(8) << v[0] << "d " << v[1] << "d " 
	    << r[0] << "\" " << r[1] << "\" " << radToDeg(angle);
      }
      break;
    case Frame::SEXAGESIMAL:
      {
	char buf[64];
	parent->mapToWCSStr(center, Frame::REF, (Frame::CoordSystem)sys,
			    (Frame::CoordFormat)format, buf, 64);
	char ra[16];
	char dec[16];
	istrstream wcs(buf,64);
	wcs >> ra >> dec;
	if (dec[0]=='+')
	  str << type << ' ' << ra << ' ' << dec+1 << ' ' 
	      << r[0] << "\" " << r[1] << "\" " << radToDeg(angle);
	else
	  str << type << ' ' << ra << ' ' << dec << ' ' 
	      << r[0] << "\" " << r[1] << "\" " << radToDeg(angle);
      }
      break;
    }
    break;
  }
}

void Ellipse::listSAOimage(ostream& str)
{
  // all coords are in image coords

  if (!(properties&INCLUDE))
    str << '-';

  Vector v = parent->map(center,Frame::REF,Frame::IMAGE);
  str << "ellipse(" << setprecision(8) << v[0] << ',' << v[1] 
      << ',' << radius[0] << ',' << radius[1] << ") ";
}
