#include <iostream.h>

#include "grid.h"
#include "util.h"
#include "frame.h"
#include "fitsimage.h"

#include <X11/Xutil.h>

extern "C" {
  #include "wcs.h"
  #include "ast.h"
  #include "grf.h"
}

typedef struct FitsStore {
  int *prj;                    /* Pointer to array of WCS projections */
  char **cdelt_name;           /* Pointers to CDELT keyword names */
  char **crota_name;           /* Pointers to CROTA keyword names */
  char **crpix_name;           /* Pointers to CRPIX keyword names */
  char **crval_name;           /* Pointers to CRVAL keyword names */
  char **ctype;                /* Pointer to array of CTYPE values */
  char **ctype_name;           /* Pointers to CTYPE keyword names */
  char **ctype_com;            /* Pointers to CTYPE keyword comments */
  char **cunit;                /* Pointer to array of CUNIT values */
  double *cdelt;               /* Pointer to array of CDELT values */
  double *crota;               /* Pointer to array of CROTA values */
  double *crpix;               /* Pointer to array of CRPIX values */
  double *crval;               /* Pointer to array of CRVAL values */
  double *pc;                  /* Pointer to the PC matrix. */
  double equinox;              /* Epoch of reference equinox */
  double latpole;              /* LATPOLE value in radians */
  double longpole;             /* LONGPOLE value in radians */
  double mjdobs;               /* Modified Julian Date of observation */
  double projp[10];  /* Projection parameters */
  int *ialt;                   /* Pointer to array of axis description cnts*/
  int axlat;                   /* Index of latitude axis */
  int axlon;                   /* Index of longitude axis */
  int julian;                  /* Is equinox value Julian? */
  int noncel;                  /* No. of non-celestial axes */
  int npar;                    /* No. of supplied projection parameters */
  int naxis;                   /* No. of axes */
  int radecsys;                /* Flag identifying frame of reference */
  int sys;                     /* Flag identifying celestial coord. sys.*/
} FitsStore;

Attribute::Attribute(Frame* p) : parent(p)
{
  style_ = SOLID;
  width_ = 1;
  size_ = 10;
  font_ = DEFAULT;
  tkfont_ = Tk_GetFont(parent->interp, parent->tkwin, "system 10 normal");
  colour_ = WHITE;

  setColour(colour_);
}

Attribute::~Attribute()
{
  if (tkfont_)
    Tk_FreeFont(tkfont_);

  if (colorName_)
    delete colorName_;
}

void Attribute::setStyle(double v)
{
  switch ((int)v) {
  case SOLID:
  case DASH:
    break;
  default:
    return;
  }

  style_ = (Style)v;
}

void Attribute::setWidth(double v)
{
  if (v >= 1)
    width_ = (int)v;
}

void Attribute::setSize(double v)
{
  if (v >= 1) {
    size_ = (int)v;
    initFont();
  }
}

void Attribute::setFont(double v)
{
  switch ((Font)v) {
  case DEFAULT:
  case TIMES:
  case HELVETICA:
  case SYMBOL:
  case COURIER:

  case DEFAULTBOLD:
  case TIMESBOLD:
  case HELVETICABOLD:
  case SYMBOLBOLD:
  case COURIERBOLD:

  case DEFAULTITALIC:
  case TIMESITALIC:
  case HELVETICAITALIC:
  case SYMBOLITALIC:
  case COURIERITALIC:
    font_ = (Font)v;
    initFont();
    return;

  default:
    return;
  }
}

void Attribute::initFont()
{
  if (tkfont_)
    Tk_FreeFont(tkfont_);
  tkfont_ = NULL;

  char buf[256];
  ostrstream str(buf,256);

  switch (font_) {
  case DEFAULT:
    str << "system " << size_ << " normal" << ends;
    break;
  case TIMES:
    str << "times " << size_ << " normal" << ends;
    break;
  case HELVETICA:
    str << "helvetica " << size_ << " normal" << ends;
    break;
  case SYMBOL:
    str << "symbol " << size_ << " normal" << ends;
    break;
  case COURIER:
    str << "courier " << size_ << " normal" << ends;
    break;

  case DEFAULTBOLD:
    str << "system " << size_ << " bold" << ends;
    break;
  case TIMESBOLD:
    str << "times " << size_ << " bold" << ends;
    break;
  case HELVETICABOLD:
    str << "helvetica " << size_ << " bold" << ends;
    break;
  case SYMBOLBOLD:
    str << "symbol " << size_ << " bold" << ends;
    break;
  case COURIERBOLD:
    str << "courier " << size_ << " bold" << ends;
    break;

  case DEFAULTITALIC:
    str << "system " << size_ << " italic" << ends;
    break;
  case TIMESITALIC:
    str << "times " << size_ << " italic" << ends;
    break;
  case HELVETICAITALIC:
    str << "helvetica " << size_ << " italic" << ends;
    break;
  case SYMBOLITALIC:
    str << "symbol " << size_ << " italic" << ends;
    break;
  case COURIERITALIC:
    str << "courier " << size_ << " italic" << ends;
    break;
  default:
    return;
  }

  tkfont_ = Tk_GetFont(parent->interp, parent->tkwin, buf);
}

void Attribute::setColour(double v)
{
  switch ((int)v) {
  case BLACK:
    colorName_ = dupstr("black");
    break;
  case WHITE:
    colorName_ = dupstr("white");
    break;
  case RED:
    colorName_ = dupstr("red");
    break;
  case GREEN:
    colorName_ = dupstr("green");
    break;
  case BLUE:
    colorName_ = dupstr("blue");
    break;
  case CYAN:
    colorName_ = dupstr("cyan");
    break;
  case MAGENTA:
    colorName_ = dupstr("magenta");
    break;
  case YELLOW:
    colorName_ = dupstr("yellow");
    break;
  default:
    return;
  }

  color_ = parent->getColor(colorName_);
  colour_ = (Colour)v;
}

Grid* gr = NULL;

Grid::Grid(Frame* p, int sys, const char* o, int m) 
  : parent(p), system_(sys), gridMode(m)
{
  option_ = dupstr(o);

  line = new Attribute(parent);
  mark = new Attribute(parent);
  text = new Attribute(parent);
  psMode = Frame::RGB;
  fits = NULL;
}

Grid::~Grid()
{
  if (option_)
    delete option_;

  if (line)
    delete line;
  if (mark)
    delete mark;
  if (text)
    delete text;
}

void Grid::render()
{
  FitsImage* ptr = parent->fits;
  while (ptr) {
    fits = ptr;
    doit(X11);
    ptr = ptr->next();
  }
}

void Grid::ps(int mode)
{
  psMode = mode;

  FitsImage* ptr = parent->fits;
  while (ptr) {
    fits = ptr;
    doit(PS);
    ptr = ptr->next();
  }
}

void Grid::doit(RenderMode rm)
{
  int width = fits->width();
  int height = fits->height();

  astClearStatus; // just to make sure
  astBegin; // start memory management
  AstFrameSet* frameSet = NULL;

  switch (system_) {
  case Frame::EQUATORIAL_FK4:
  case Frame::EQUATORIAL_FK5:
  case Frame::ICRS:
  case Frame::GALACTIC:
  case Frame::ECLIPTIC:
    {
      if (!fits->hasWCSEqu()) {
	cerr << "Error: Grid- no wcs available" << endl;
	astEnd;
	return;
      }

      WorldCoor* wcs = fits->getWCS();
      if (wcs->prjcode==WCS_DSS) {
	frameSet = (AstFrameSet*)wcsdss2ast(wcs,
					    !strncmp(wcs->radecsys,"FK5",3),
					    wcs->equinox, wcs->epoch);
      }
      else if ((wcs->prjcode>=WCS_AZP && wcs->prjcode<=WCS_NCP) ||
	       (wcs->prjcode==WCS_PLT) ||
	       (wcs->prjcode==WCS_TNX)) {
	FitsStore* store = wcs2Store();
	if (store) {
	  frameSet = (AstFrameSet*)wcs2ast(store);
	  deleteStore(store);
	}
      }

      // frameSet = (AstFrameSet*)ast2Frame();
      // astShow(frameSet);

      if (frameSet == AST__NULL) {
	cerr << "Error: Grid- unable to decode wcs for coordinate grid" 
	     << endl;
	astEnd;
	return;
      }

      // set desired WCS system

      switch (system_) {
      case Frame::EQUATORIAL_FK4:
	astSet(frameSet, "System=FK4, Equinox=B1950.0");
	break;
      case Frame::EQUATORIAL_FK5:
	astSet(frameSet, "System=FK5, Equinox=J2000.0");
	break;
      case Frame::ICRS:
	// ast does not support icrs
	astSet(frameSet, "System=FK5, Equinox=J2000.0");
	break;
      case Frame::GALACTIC:
	astSet(frameSet, "System=Galactic");
	break;
      case Frame::ECLIPTIC:
	astSet(frameSet, "System=Ecliptic");
	break;
      }
    }
    break;

  case Frame::LINEAR:
    {
      if (!fits->hasWCSLinear()) {
	cerr << "Error: Grid- no linear wcs available" << endl;
	astEnd;
	return;
      }

      WorldCoor* wcs = fits->getWCS();
      Matrix m(wcs->cdelt[0],0,0,wcs->cdelt[1],
	       wcs->crval[0]-(wcs->cdelt[0]*wcs->crpix[0]),
	       wcs->crval[1]-(wcs->cdelt[1]*wcs->crpix[1]));

      Vector imgll(1,1);
      Vector imgur = Vector(width, height);
      Vector linll = imgll * m;
      Vector linur = imgur * m;

      frameSet = astFrameSet(astFrame(2, ""), "");
      astAddFrame(frameSet, 1, 
		  astWinMap(2, imgll.v, imgur.v, linll.v, linur.v, ""),
		  astFrame(2, ""));
    }
    break;

  case Frame::DETECTOR:
    {
      Matrix m = fits->getImageToPhysical() * fits->getPhysicalToDetector();
      Vector imgll(1,1);
      Vector imgur = Vector(width, height);
      Vector detll = imgll * m;
      Vector detur = imgur * m;

      frameSet = astFrameSet(astFrame(2, ""), "");
      astAddFrame(frameSet, 1, 
		  astWinMap(2, imgll.v, imgur.v, detll.v, detur.v, ""),
		  astFrame(2, ""));
    }
    break;

  case Frame::AMPLIFIER:
    {
      Matrix m = fits->getImageToPhysical() * fits->getPhysicalToAmplifier();
      Vector imgll(1,1);
      Vector imgur = Vector(width, height);
      Vector ampll = imgll * m;
      Vector ampur = imgur * m;

      frameSet = astFrameSet(astFrame(2, ""), "");
      astAddFrame(frameSet, 1, 
		  astWinMap(2, imgll.v, imgur.v, ampll.v, ampur.v, ""),
		  astFrame(2, ""));
    }
    break;

  case Frame::PHYSICAL:
    {
      Vector imgll(1,1);
      Vector imgur = Vector(width, height);
      Vector phyll = imgll * fits->getImageToPhysical();
      Vector phyur = imgur * fits->getImageToPhysical();

      frameSet = astFrameSet(astFrame(2, ""), "");
      astAddFrame(frameSet, 1, 
		  astWinMap(2, imgll.v, imgur.v, phyll.v, phyur.v, ""),
		  astFrame(2, ""));
    }
    break;

  case Frame::IMAGE:
    frameSet = astFrameSet(astFrame(2,""),"");
    break;
  }

  astSet(frameSet,"Title=%s, Label(1)=%s, Label(2)=%s", " ", " ", " ");

  // create astPlot

  float gbox[4];
  double pbox[4];

  if (gridMode || (parent->fitsCount>1)) {
    gbox[0] = pbox[0] = 1;
    gbox[1] = pbox[1] = 1;
    gbox[2] = pbox[2] = width;
    gbox[3] = pbox[3] = height;
  }
  else {
    Matrix m = fits->getWidgetToImage();
    Vector ll = Vector(0,0) * m;
    Vector lr = Vector(parent->options->width,0) * m;
    Vector ur = Vector(parent->options->width, parent->options->height) * m;
    Vector ul = Vector(0, parent->options->height) * m;

    BBox bb(ll,ll);
    bb.bound(lr);
    bb.bound(ur);
    bb.bound(ul);

    gbox[0] = pbox[0] = (int)bb.ll[0];
    gbox[1] = pbox[1] = (int)bb.ll[1];
    gbox[2] = pbox[2] = (int)bb.ur[0];
    gbox[3] = pbox[3] = (int)bb.ur[1];
  }

  AstPlot* plot = astPlot(frameSet, gbox, pbox, option_);

  // and now create astGrid

  gr = this;
  renderMode = rm;
  astGrid(plot);

  astEnd; // now, clean up memory
}

// Generic

void Grid::gAttr(int which, double value, double* old, int prim)
{
  if (value == AST__BAD) {
    if (old)
      *old = 0;
    return;
  }

  Attribute* attr;
  switch (prim) {
  case GRF__TEXT:
    attr = text;
    break;
  case GRF__LINE:
    attr = line;
    break;
  case GRF__MARK:
    attr = mark;
    break;
  }

  switch (which) {
  case GRF__STYLE:
    if (old)
      *old = attr->style();
    attr->setStyle(value);
    break;
  case GRF__WIDTH:
    if (old)
      *old = attr->width();
    attr->setWidth(value);
    break;
  case GRF__SIZE:
    if (old)
      *old = attr->size();
    attr->setSize(value);
    break;
  case GRF__FONT:
    if (old)
      *old = attr->font();
    attr->setFont(value);
    break;
  case GRF__COLOUR:
    if (old)
      *old = attr->colour();
    attr->setColour(value);
    break;
  }
}

void Grid::gQch(float* chv, float* chh)
{
  //  cerr << "gQch" << endl;
  if (text->tkfont()) {
    Tk_FontMetrics metrics;
    Tk_GetFontMetrics(text->tkfont(), &metrics);

    *chv = (float)metrics.linespace;
    *chh = (float)metrics.linespace;
  }
  else
    *chv = *chh = 0;
}

void Grid::gTxExt(const char* txt, float x, float y, const char* just,
		  float upx, float upy, float* xb, float* yb)
{
  if (txt && text->tkfont()) {
    Tk_FontMetrics metrics;
    Tk_GetFontMetrics(text->tkfont(), &metrics);
    int width = Tk_TextWidth(text->tkfont(), txt, strlen(txt));

    Matrix m1,m2;
    if (just) {
      switch (just[0]) {
      case 'T':
	m1 = Translate(0,metrics.ascent);
	break;
      case 'C':
	break;
      case 'B':
	m1 = Translate(0,metrics.descent);
	break;
      }

      switch (just[1]) {
      case 'L':
	break;
      case 'C':
	m2 = Translate(-width/2,0);
	break;
      case 'R':
	m2 = Translate(-width,0);
	break;
      }
    }

    Vector c = Vector(x,y) * fits->getImageToWidget() *  m1 * m2;

    BBox bb = BBox(c*Translate(-width/2,-metrics.descent), 
		   c*Translate(width/2,metrics.ascent));

    xb[0] = xb[1] = bb.ll[0];
    xb[2] = xb[3] = bb.ur[0];

    yb[0] = yb[3] = bb.ll[1];
    yb[1] = yb[2] = bb.ur[1];
  }
  else {
    xb[0] = xb[1] = xb[2] = xb[3] = 0;
    yb[0] = yb[1] = yb[2] = yb[3] = 0;
  }
}

// X11 Render functions

void Grid::x11Line(int n, const float* x, const float* y)
{
  if (n<2 || !x || !y)
    return;

  XSetForeground(parent->display, parent->gridGC, line->color());
  switch (line->style()) {
  case Attribute::SOLID:
    XSetLineAttributes(parent->display, parent->gridGC, line->width(), 
		       LineSolid, CapButt, JoinMiter);
    break;
  case Attribute::DASH:
    XSetLineAttributes(parent->display, parent->gridGC, line->width(), 
		       LineOnOffDash, CapButt, JoinMiter);
    char dlist[] = {8,3};
    XSetDashes(parent->display, parent->gridGC, 0, dlist, 2);
    break;
  }

  Matrix m = fits->getImageToWidget();

  Vector s = Vector(x[0],y[0]) * m;
  Vector e;
  for (int i=0; i<n; i++) {
    e = Vector(x[i],y[i]) * m;
    XDrawLine(parent->display, parent->basePixmap, parent->gridGC, 
	      (int)s[0],(int)s[1],(int)e[0],(int)e[1]);
    s = e;
  }
}

void Grid::x11Text(const char* txt, float x, float y, 
		   const char* just, float upx, float upy)
{
  if (txt && txt[0] && text->tkfont() && just) {
    XSetFont(parent->display, parent->gridGC, Tk_FontId(text->tkfont()));
    XSetForeground(parent->display, parent->gridGC, text->color());

    Vector v = Vector(x,y) * fits->getImageToWidget();

    int flip = 0;
    double angle = calcTextAngle(Vector(upx,upy).angle(), &flip);

    if ((angle>-.01 && angle<+.01) && !flip)
      DrawString(txt, v, just);
    else
      DrawRotString(txt, v, just, angle, flip);
  }
}

void Grid::DrawString(const char* txt, Vector v, const char* j)
{
  Tk_FontMetrics metrics;
  Tk_GetFontMetrics(text->tkfont(), &metrics);
  int width = Tk_TextWidth(text->tkfont(), txt, strlen(txt));

  switch (j[0]) {
  case 'T':
    v[1] += metrics.ascent;
    break;
  case 'C':
    break;
  case 'B':
    v[1] -= metrics.descent;
    break;
  }

  switch (j[1]) {
  case 'L':
    break;
  case 'C':
    v[0] -= width/2.;
    break;
  case 'R':
    v[0] -= width;
    break;
  }

  XDrawString(parent->display, parent->basePixmap, parent->gridGC,
	      (int)v[0], (int)v[1], txt, strlen(txt));
}

void Grid::DrawRotString(const char* txt, Vector v, 
			 const char* just, double angle, int flip)
{
  Display* display = parent->display;
  GC gc = parent->gridGC;
  Pixmap base = parent->basePixmap;

  Tk_FontMetrics metrics;
  Tk_GetFontMetrics(text->tkfont(), &metrics);
  int width = Tk_TextWidth(text->tkfont(), txt, strlen(txt));

  // calculate bounding box
  int srcw = width+4;
  int srch = metrics.ascent+metrics.descent+2;
  Vector corner = (Vector(srcw,srch)-Vector(1,1))/2;
  BBox bb;
  bb.bound(-corner * Rotate(angle));
  bb.bound( corner * Rotate(angle));
  bb.bound(Vector( corner[0],-corner[1]) * Rotate(angle));
  bb.bound(Vector(-corner[0], corner[1]) * Rotate(angle));
  Vector dstsize = bb.size();
  int dstw = (int)dstsize[0];
  int dsth = (int)dstsize[1];

  // create bitmap
  Pixmap srcbmap = XCreatePixmap(display, base, srcw, srch, 1);
  Pixmap dstbmap = XCreatePixmap(display, base, dstw, dsth, 1);

  // create bitmap gc
  GC bgc = XCreateGC(display, srcbmap, 0L, NULL);
  XGCValues values;
  XGetGCValues(display,gc,GCFont,&values);
  XChangeGC(display,bgc,GCFont,&values);
  XSetBackground(display, bgc, 0);

  // clear bitmaps
  XSetForeground(display, bgc, 0);
  XFillRectangle(display, srcbmap, bgc, 0, 0, srcw, srch);
  XFillRectangle(display, dstbmap, bgc, 0, 0, dstw, dsth);

  XSetForeground(display, bgc, 1);
  XDrawString(display, srcbmap, bgc, 2, metrics.ascent, txt, strlen(txt));

  // create xmaps
  XImage* srcxmap = XGetImage(display, srcbmap, 0, 0, srcw, srch,
			    AllPlanes, XYPixmap);
  XImage* dstxmap = XGetImage(display, dstbmap, 0, 0, dstw, dsth,
			    AllPlanes, XYPixmap);

  // rotate text from srcxmap to dstxmap
  {
    char* src = srcxmap->data;
    char* dst = dstxmap->data;
    int srcl = srcxmap->bytes_per_line;
    int dstl = dstxmap->bytes_per_line;
    Matrix m = 
      Translate(-Vector(srcw,srch)/2) *
      Rotate(angle) *
      Translate(Vector(dstw,dsth)/2);
    Matrix mm = m.invert();

    for (int j=0; j<dsth; j++)
      for (int i=0; i<dstw; i++) {

	Vector s = Vector(i,j) * mm;
	if (s[0]>=0 && s[0]<srcw && s[1]>=0 && s[1]<srch) {

	  int soff = ((int)s[1])*srcl+(((int)s[0])/8);
	  int sshift = ((int)s[0])%8;

	  if (parent->baseXImage->byte_order) {
	    // NOTE: ximage gave us the bitmap with bytes MSB, so we must
	    // 128>>shift instead of 1<<shift.
	    if (*(src+soff) & (128>>sshift)) {
	      int doff = j*dstl+(i/8);
	      int dshift = i%8;
	      *(dst+doff) |= 128>>dshift;
	    }
	  }
	  else {
	    if (*(src+soff) & (1<<sshift)) {
	      int doff = j*dstl+(i/8);
	      int dshift = i%8;
	      *(dst+doff) |= 1<<dshift;
	    }
	  }
	}
      }
  }

  // put rotated text into dstbmap
  XPutImage(display, dstbmap, bgc, dstxmap, 0, 0, 0, 0, dstw, dsth);

  // create new GC
  GC ggc = XCreateGC(display, base, 0L, NULL);
  XCopyGC(display, gc, GCForeground, ggc);
  XSetStipple(display, ggc, dstbmap);
  XSetFillStyle(display, ggc, FillStippled);

  // calculate rect location
  Matrix m1;
  switch (just[0]) {
  case 'T':
    m1 = !flip ? Translate(0, corner[1]) : Translate(0, -corner[1]);
    break;
  case 'C':
    break;
  case 'B':
    m1 = !flip ? Translate(0, -corner[1]) : Translate(0, corner[1]);
    break;
  }

  Matrix m2;
  switch (just[1]) {
  case 'L':
    m2 = Translate(-corner[0], 0);
    break;
  case 'C':
    break;
  case 'R':
    m2 = Translate(corner[0], 0);
    break;
  }

  Matrix m = m1 * m2 * Rotate(angle) * Translate(v);
  BBox vv(v,v); // initialize to position v
  vv.bound(-corner * m);
  vv.bound( corner * m);
  vv.bound(Vector( corner[0],-corner[1]) * m);
  vv.bound(Vector(-corner[0], corner[1]) * m);

  Vector t = vv.ll;

  // stencil in the text
  XSetTSOrigin(display, ggc, (int)t[0], (int)t[1]);
  XFillRectangle(display, base, ggc, (int)t[0], (int)t[1], dstw, dsth);

  // clean up
  XFreePixmap(display, srcbmap);
  XFreePixmap(display, dstbmap);
  XFreeGC(display, bgc);
  XFreeGC(display, ggc);
  XDestroyImage(srcxmap);
  XDestroyImage(dstxmap);
}

// PS Render functions

void Grid::psLine(int n, const float* x, const float* y)
{
  if (n<2 || !x || !y)
    return;

  char buf[256];
  Matrix m = fits->getImageToCanvas();

  psColor(line);
  psWidth(line);
  psStyle(line);

  {
    ostrstream str(buf,256);
    Vector v = Vector(x[0],y[0]) * m;
    str << "newpath " << endl
	<< v.TkCanvasPs(parent->canvas) << " moveto" << endl << ends;
    Tcl_AppendResult(parent->interp, buf, NULL);
  }

  for (int i=1; i<n; i++) {
    ostrstream str(buf,256);
    Vector v = Vector(x[i],y[i]) * m;
    str << v.TkCanvasPs(parent->canvas) << " lineto" << endl << ends;
    Tcl_AppendResult(parent->interp, buf, NULL);
  }

  {
    ostrstream str(buf,256);
    str << "stroke" << endl << ends;
    Tcl_AppendResult(parent->interp, buf, NULL);
  }
}

void Grid::psText(const char* txt, float x, float y, 
		  const char* just, float upx, float upy)
{
  if (txt && txt[0] && text->tkfont() && just) {
    int flip = 0;
    double angle = calcTextAngle(Vector(upx,upy).angle(),&flip);

    Tk_FontMetrics metrics;
    Tk_GetFontMetrics(text->tkfont(), &metrics);
    int width = Tk_TextWidth(text->tkfont(), txt, strlen(txt));

    Matrix m1;
    switch (just[0]) {
    case 'T':
      m1 = !flip ? Translate(0,metrics.ascent) : Translate(0,-metrics.descent);
      break;
    case 'C':
      break;
    case 'B':
      m1 = !flip ? Translate(0,-metrics.descent) : Translate(0,metrics.ascent);
      break;
    }

    Matrix m2;
    switch (just[1]) {
    case 'L':
      break;
    case 'C':
      m2 = Translate(-width/2.,0);
      break;
    case 'R':
      m2 = Translate(-width,0);
      break;
    }

    Vector v = Vector(x,y) * fits->getImageToCanvas();
    Matrix m = Translate(-v) * m1 * m2 * Rotate(angle) * Translate(v);
    Vector t = v * m;
    
    Tcl_DString psFont;
    Tcl_DStringInit(&psFont);
    int psSize = Tk_PostscriptFontName(text->tkfont(), &psFont);

    char buf[512];
    ostrstream str(buf,512);
    str << '/' << Tcl_DStringValue(&psFont) 
	<< " findfont " << psSize << " scalefont setfont" << endl;

    Tcl_DStringFree(&psFont);

    psColor(text);

    str << "gsave " 
	<< t.TkCanvasPs(parent->canvas) << " moveto" << endl
      	<< radToDeg(angle) << " rotate " 
	<< '(' << txt << ')' << " show"
	<< " grestore"
	<< endl << ends;

    Tcl_AppendResult(parent->interp, buf, NULL);
  }
}

// PS Support

void Grid::psColor(Attribute* attr)
{
  char buf[256];

  ostrstream str(buf,256);
  switch ((Frame::PSColorSpace)psMode) {
  case Frame::BW:
  case Frame::GRAY:
    psColorGray(attr->colorName(), str);
    str << " setgray";
    break;
  case Frame::RGB:
    psColorRGB(attr->colorName(), str);
    str << " setrgbcolor";
    break;
  case Frame::CMYK:
    psColorCMYK(attr->colorName(), str);
    str << " setcmykcolor";
    break;
  }
  str << endl << ends;

  Tcl_AppendResult(parent->interp, buf, NULL);
}

void Grid::psWidth(Attribute* attr)
{
  char buf[256];
  ostrstream str(buf,256);
  str << attr->width() << " setlinewidth" << endl << ends;
  Tcl_AppendResult(parent->interp, buf, NULL);
}

void Grid::psStyle(Attribute* attr)
{
  char buf[256];
  ostrstream str(buf,256);

  switch (attr->style()) {
  case Attribute::SOLID:
    str << "[] 0 setdash" << endl << ends;
    break;
  case Attribute::DASH:
    str << "[8 3] 0 setdash" << endl << ends;
    break;
  }

  Tcl_AppendResult(parent->interp, buf, NULL);
}

// AST support

void* Grid::ast2Frame() 
{
  // read FITS header into astChannel

  astClearStatus;
  AstFitsChan* chan = astFitsChan(NULL, NULL, "");
  if (!astOK || chan == AST__NULL)
    return NULL;

  char* header = fits->getHeader();
  int ncards = fits->getNCards();
  for (int i=0; i<ncards; i++) {
    char buf[81];
    strncpy(buf,header+(i*80),80);
    buf[80] = '\0';
    // sometimes, we get a bad parse, just ignore
    if (!astOK)
      astClearStatus;
    astPutFits(chan, buf, 0);
  }

  // parse FITS header for WCS

  if (!astOK)
    astClearStatus;
  astClear(chan, "Card");
  AstFrameSet* frameSet = (AstFrameSet*)astRead(chan);
  if (!astOK || frameSet == AST__NULL || 
      strcmp(astGetC(frameSet,"Class"), "FrameSet"))
    return NULL;

  return frameSet;
}

extern "C" {
  double slaEpb2d(double);
}

FitsStore* Grid::wcs2Store()
{
  WorldCoor* wcs = fits->getWCS();

  // make sure we have availd projection
  if (wcs->prjcode<WCS_AZP || wcs->prjcode==WCS_DSS)
    return NULL;

  FitsStore* store = new FitsStore;

  store->naxis = 2;
  store->noncel = 0;
  if (!wcs->coorflip) {
    store->axlat = 1;
    store->axlon = 0;
  }
  else {
    store->axlat = 0;
    store->axlon = 1;
  }

  // crval is in radians
  store->crval = new double[2];
  store->crval[0] = wcs->crval[0] * AST__DD2R;
  store->crval[1] = wcs->crval[1] * AST__DD2R;
  store->crval_name = new char*[2];
  store->crval_name[0] = dupstr("CRVAL1");
  store->crval_name[1] = dupstr("CRVAL2");

  store->crpix = wcs->crpix;
  store->crpix_name = new char*[2];
  store->crpix_name[0] = dupstr("CRPIX1");
  store->crpix_name[1] = dupstr("CRPIX2");

  store->ctype = new char*[2];
  if (!strncmp(wcs->ctype[0],"RA---TNX",8)) {
    store->ctype[0] = dupstr("RA---TAN");
    store->ctype[1] = dupstr("DEC--TAN");
  }
  else if (!strncmp(wcs->ctype[0],"DEC--TNX",8)) {
    store->ctype[0] = dupstr("DEC--TAN");
    store->ctype[1] = dupstr("RA---TAN");
  }
  else {
    store->ctype[0] = dupstr(wcs->ctype[0]);
    store->ctype[1] = dupstr(wcs->ctype[1]);
  }

  store->ctype_name = new char*[2];
  store->ctype_name[0] = dupstr("CTYPE1");
  store->ctype_name[1] = dupstr("CTYPE2");
  store->ctype_com = new char*[2];
  store->ctype_com[0] = NULL;
  store->ctype_com[1] = NULL;

  // cdelt is in radians
  store->cdelt = new double[2];
  store->cdelt[0] = wcs->cdelt[0] * AST__DD2R;
  store->cdelt[1] = wcs->cdelt[1] * AST__DD2R;
  store->cdelt_name = new char*[2];
  store->cdelt_name[0] = dupstr("CDELT1");
  store->cdelt_name[1] = dupstr("CDELT2");

  store->crota = new double[2];
  store->crota[0] = wcs->coorflip ? AST__BAD : wcs->rot;
  store->crota[1] = wcs->coorflip ? wcs->rot : AST__BAD;
  store->crota_name = new char*[2];
  store->crota_name[0] = NULL;
  store->crota_name[1] = NULL;

  store->cunit = new char*[2];
  store->cunit[0] = NULL;
  store->cunit[1] = NULL;

  // will be set later
  store->prj = new int[2];
  store->prj[0] = AST__WCSBAD;
  store->prj[1] = AST__WCSBAD;

  store->ialt = new int[2];
  store->ialt[0] = 1;
  store->ialt[1] = 2;

  switch (wcs->syswcs) {
  case WCS_J2000:
  case WCS_B1950:
    store->sys = 1; // RADEC
    break;
  case WCS_GALACTIC:
    store->sys = 3; // GALAC
    break;
  case WCS_ECLIPTIC:
    store->sys = 2; // ECLIP
    break;
  default:
    // cerr << "Error: Grid- Can't convert wcssub SYSWCS to ast SYS" << endl;
    store->sys = 0; // NOCEL
  }

  {
    int yes = 0;
    for (int j=0; j<4; j++)
      if (wcs->cd[j]!=0)
	yes = 1;

    if (yes) {
      store->pc = new double[4];
      for (int i=0; i<4; i++)
	store->pc[i] = wcs->cd[i];
    }
    else
      store->pc = NULL;
  }

  {
    store->npar = 0;
    for (int i=0; i<10; i++) {
      if (wcs->projp[i] != 0) {
	store->projp[i] = wcs->projp[i];
	store->npar++;
      }
      else 
	store->projp[i] = AST__BAD;
    }
  }

  store->longpole = AST__BAD;
  store->latpole = AST__BAD;

  if (!strncmp(wcs->radecsys,"FK5",3)) {
    store->radecsys = 3; // FK5
  }
  else if (!strncmp(wcs->radecsys,"ICRS",4)) {
    store->radecsys = 3; // FK5
  }
  else if (!strncmp(wcs->radecsys,"FK4-",4)) {
    store->radecsys = 2; // FK4-NO-E
  }
  else if (!strncmp(wcs->radecsys,"FK4",3)) {
    store->radecsys = 1; // FK4
  }
  else if (!strncmp(wcs->radecsys,"GAPP",4)) {
    store->radecsys = 4; // GAPPT
  }
  else {
    store->radecsys = 0; // NORADEC
  }

  store->equinox = wcs->equinox;
  if( store->equinox < 1984.0 )
    store->julian = 0;
  else
    store->julian = 1;

  store->mjdobs = slaEpb2d(wcs->epoch);

  return store;
}

void Grid::deleteStore(FitsStore* store)
{
  delete store->crval;
  delete store->crval_name[0];
  delete store->crval_name[1];
  delete store->crval_name;

  delete store->crpix_name[0];
  delete store->crpix_name[1];
  delete store->crpix_name;

  delete store->ctype[0];
  delete store->ctype[1];
  delete store->ctype;

  delete store->ctype_name[0];
  delete store->ctype_name[1];
  delete store->ctype_name;

  delete store->ctype_com;

  delete store->cdelt;
  delete store->cdelt_name[0];
  delete store->cdelt_name[1];
  delete store->cdelt_name;

  delete store->crota;
  delete store->crota_name;

  delete store->cunit;

  delete store->prj;
  delete store->ialt;
  delete store->pc;

  delete store;
}

double Grid::calcTextAngle(double a, int* flip)
{
  if (flip)
    *flip = 0;

  // our angle is 90 off from ast's
  double r = a - M_PI/2;

  // and add any image rotation
  r += parent->getRotation();

  // normalize
  if (r>0)
    while (r>2*M_PI)
      r -= 2*M_PI;
  else 
    while (r<0)
      r += 2*M_PI;

  // and flip if upside down
  // we only want to do this if the number is really, really upside down
  // you might have the case of a curved vertical line
  if (r > M_PI/2+.1 && r < 3*M_PI/2-.1) {
    r -= M_PI;
    *flip = 1;
  }

  return r;
}
